/**
 * @addtogroup ChefTest
 * @{
 * @file ChefTest/CcLibs_Option.cc
 * @author Massimiliano Pagani
 *
 */

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ChefFun/Option.hh>
#include <string>
#include <gtest/gtest.h>

using namespace ChefFun;

TEST( ChefFun_Option,  copy_constructors )
{
    auto a = Option<int>::Some(3);
    Option<int> b{a};

    ASSERT_TRUE( b.isDefined() );
    ASSERT_EQ( b.get(), 3 );

    auto c = Option<int>::None();
    Option<int> d{c};
    ASSERT_TRUE( c.isEmpty() );
}

TEST( ChefFun_Option,  base_test )
{
    typedef Option<int> OptionInt;

    OptionInt a = None<int>();
    ASSERT_TRUE( a.isEmpty() );

    bool caught=false;
    try
    {
        int x = a.move();
        (void)x;
        assert( false );
    }
    catch( ... )
    {
        caught = true;
    }
    assert( caught );

    a = Some(3);
    ASSERT_TRUE( a.isDefined() );
    OptionInt b{ OptionInt::Some(3)};
    ASSERT_TRUE( b.isDefined() );
    ASSERT_EQ( None<int>().getOrElse(3), 3 );
    ASSERT_EQ( OptionInt::Some(42).getOrElse(3), 42 );

    int c = 1+b.get();
    ASSERT_EQ( c, 1+3 );
}

static int f( std::string const&  x )
{
    return x.length();
}

TEST( ChefFun_Option,  vararg_constructor )
{
    class Test
    {
        public:
            Test( int x, int y ) : m_x{x},m_y{y} {}
            int m_x;
            int m_y;
    };

    auto o = Option<Test>::Some( 4, 2 );
    ASSERT_TRUE( o.isDefined() );
    ASSERT_TRUE( o.get().m_x == 4 );
    ASSERT_TRUE( o.get().m_y == 2 );
}

TEST( ChefFun_Option,  map_test )
{
    Option<std::string> test = Option<std::string>::Some( std::string{"abc"});
    Option<int> result = test.map( f );
    ASSERT_TRUE( result.isDefined() );
    ASSERT_EQ( result.get(), 3 );
}

TEST( ChefFun_Option,  assignment )
{
    Option<int> a{ Option<int>::Some(3) };
    Option<int> b{ Option<int>::Some(2) };

    a = b;
    ASSERT_TRUE( a.isDefined() );
    ASSERT_TRUE( b.isDefined() );
    ASSERT_TRUE( a.get() == 2 );
    ASSERT_TRUE( b.get() == 2 );
    ASSERT_TRUE( a == b );

    Option<int> c{ Option<int>::None() };
    a = c;
    ASSERT_TRUE( a.isEmpty() );
    ASSERT_TRUE( a == c );

    a = b;
    ASSERT_TRUE( a.isDefined() );
    ASSERT_TRUE( b.isDefined() );
    ASSERT_TRUE( a.get() == 2 );
    ASSERT_TRUE( b.get() == 2 );
    ASSERT_TRUE( a == b );

    a = c;
    // both a and c are None
    c = a;
    ASSERT_TRUE( a.isEmpty() );
    ASSERT_TRUE( c.isEmpty() );

}

TEST( ChefFun_Option,  foreach )
{
    auto a = Option<int>::Some(3);
    auto b = Option<int>::None();

    bool isCalled = false;
    a.foreach( [&isCalled]( int x ){ isCalled = true; });
    ASSERT_TRUE( isCalled );

    isCalled = false;
    b.foreach( [&isCalled]( int x ){ isCalled = true; });
    ASSERT_TRUE( !isCalled );
}

TEST( ChefFun_Option,  begin_end )
{
    auto a = Option<int>::Some(3);
    auto b = Option<int>::None();

    bool isExecuted = false;
    for( auto x : a )
    {
        isExecuted = true;
        ASSERT_TRUE( x == 3 );
    }
    ASSERT_TRUE( isExecuted );

    isExecuted = false;
    for( auto x : b )
    {
        (void)x;
        isExecuted = true;
    }
    ASSERT_TRUE( !isExecuted );
}

TEST( ChefFun_Option,  flatten )
{
    auto a = Option<Option<int>>::Some(Option<int>::Some(3));
    auto b = Option<Option<int>>::Some(Option<int>::None());
    auto c = Option<Option<int>>::None();

    ASSERT_TRUE( a.flatten() == Option<int>::Some(3) );
    ASSERT_TRUE( b.flatten() == Option<int>::None() );
    ASSERT_TRUE( c.flatten() == Option<int>::None() );
}

TEST( ChefFun_Option,  pattern_matching_functional )
{
    auto a = Option<int>::Some(3);
    auto b = Option<int>::None();

    int result0 = a
        .matchSome( []( auto&& n ){ return n*2; } )
        .matchNone( [](){ return 0; } );

    ASSERT_EQ( result0, 6 );

    int result1 = b
        .matchSome( []( auto&& n ){ return n*2; } )
        .matchNone( [](){ return 0; } );

    ASSERT_EQ( result1, 0 );
}

TEST( ChefFun_Option,  pattern_matching_imperative )
{
    auto a = Option<int>::Some(3);
    auto b = Option<int>::None();

    int result{};

    a
        .matchSome( [&result]( auto&& n ){ result = n*2; } )
        .matchNone( [&result](){ result = 0; } );

    ASSERT_EQ( result, 6 );

    b
        .matchSome( [&result]( auto&& n ){ result = n*2; } )
        .matchNone( [&result](){ result = 0; } );

    ASSERT_EQ( result, 0 );
}


///@}
